"""
    Mapping helpers
"""

import numpy as np
import pandas as pd
import scanpy as sc
import torch
import logging

from scipy.sparse.csc import csc_matrix
from scipy.sparse.csr import csr_matrix

from . import mapping_optimizer as mo
from . import utils as ut


def pp_adatas(adata_1, adata_2, genes=None):
    """
    Pre-process AnnDatas so that they can be mapped. Specifically:
    - Subset the AnnDatas to `genes` (non-shared genes are removed).
    - Re-order genes in `adata_2` so that they are consistent with those in `adata_1`.
    :param adata_1:
    :param adata_2:
    :param genes:
    List of genes to use. If `None`, all genes are used.
    :return:
    """
    adata_1 = adata_1.copy()
    adata_2 = adata_2.copy()
    adata_1.var_names_make_unique()
    adata_2.var_names_make_unique()
    
    if genes is None:
        # Use all genes
        genes = adata_1.var.index.values
    else:
        genes = list(genes)
    
    # Refine `marker_genes` so that they are shared by both adatas
    genes = list(set(genes) & set(adata_1.var.index) & set(adata_2.var.index))
    logging.info(f'{len(genes)} marker genes shared by AnnDatas.')

    # Subset adatas on marker genes
    adata_1 = adata_1[:, genes]
    adata_2 = adata_2[:, genes]

    assert adata_2.var.index.equals(adata_1.var.index)
    return adata_1, adata_2


def map_cells_to_space(adata_cells, adata_space, mode='simple', adata_map=None,
                      device='cuda:0', learning_rate=0.1, num_epochs=1000):
    """
        Map single cell data (`adata_1`) on spatial data (`adata_2`). If `adata_map`
        is provided, resume from previous mapping.
        Returns a cell-by-spot AnnData containing the probability of mapping cell i on spot j.
        The `uns` field of the returned AnnData contains the training genes.
    """
    
    if adata_cells.var.index.equals(adata_space.var.index) is False:
        logging.error('Incompatible AnnDatas. Run `pp_adatas().')
        raise ValueError
    
    logging.info('Allocate tensors for mapping.')
    # Allocate tensors (AnnData matrix can be sparse or not)
    if isinstance(adata_cells.X, csc_matrix) or isinstance(adata_cells.X, csr_matrix):
        S = np.array(adata_cells.X.toarray(), dtype='float32')
    elif isinstance(adata_cells.X, np.ndarray):
        S = np.array(adata_cells.X, dtype='float32')
    else:
        X_type = type(adata_cells.X)
        logging.error('AnnData X has unrecognized type: {}'.format(X_type))
        raise NotImplementedError
    if isinstance(adata_space.X, csc_matrix) or isinstance(adata_space.X, csr_matrix):
        G = np.array(adata_space.X.toarray(), dtype='float32')
    elif isinstance(adata_space.X, np.ndarray):
        G = np.array(adata_space.X, dtype='float32')
    else:
        X_type = type(adata_space.X)
        logging.error('AnnData X has unrecognized type: {}'.format(X_type))
        raise NotImplementedError
    d = np.zeros(adata_space.n_obs)
    
    # Choose device
    device = torch.device(device)  # for gpu

    # Init hyperparameters
    if mode == 'simple':
        hyperparameters = {
            'lambda_d': 0,  # KL (ie density) term
            'lambda_g1': 1,  # gene-voxel cos sim
            'lambda_g2': 0,  # voxel-gene cos sim
            'lambda_r': 0,  # regularizer: penalize entropy
        }
    else:
        raise NotImplementedError

    # Train Tangram
    logging.info('Begin training...')
    mapper = mo.Mapper(
        S=S, G=G, d=d, device=device, adata_map=adata_map,
        **hyperparameters,
    )
    # TODO `train` should return the loss function
    mapping_matrix = mapper.train(
        learning_rate=learning_rate,
        num_epochs=num_epochs
    )

    logging.info('Saving results..')
    adata_map = sc.AnnData(X=mapping_matrix,
                           obs=adata_cells.obs.copy(),
                           var=adata_space.obs.copy())

    # Annotate cosine similarity of each training gene
    G_predicted = (adata_map.X.T @ S)
    cos_sims = []
    for v1, v2 in zip(G.T, G_predicted.T):
        norm_sq = np.linalg.norm(v1) * np.linalg.norm(v2)
        cos_sims.append((v1 @ v2) / norm_sq)
    training_genes = list(np.reshape(adata_cells.var.index.values, (-1,)))
    df_cs = pd.DataFrame(cos_sims, training_genes, columns=['train_score'])
    df_cs = df_cs.sort_values(by='train_score', ascending=False)
    adata_map.uns['train_genes_df'] = df_cs
    
    # Annotate sparsity of each training genes
    ut.annotate_gene_sparsity(adata_cells)
    ut.annotate_gene_sparsity(adata_space)
    adata_map.uns['train_genes_df']['sparsity_sc'] = adata_cells.var.sparsity
    adata_map.uns['train_genes_df']['sparsity_sp'] = adata_space.var.sparsity
    adata_map.uns['train_genes_df']['sparsity_diff'] = adata_space.var.sparsity - adata_cells.var.sparsity
    
    return adata_map






"""
    Utility functions to pre- and post-process data for Tangram.
"""
import numpy as np
import pandas as pd
from collections import defaultdict
import gzip
import pickle
import scanpy as sc

from . import mapping_utils as mu


def read_pickle(filename):
    """
        Helper to read pickle file which may be zipped or not.
        Args:
            filename (str): A valid string path.
        Returns:
            The file object.
    """
    try:
        with gzip.open(filename, 'rb') as f:
            loaded_object = pickle.load(f)
            return loaded_object
    except OSError:
        with open(filename, 'rb') as f:
            loaded_object = pickle.load(f)
            return loaded_object


def annotate_gene_sparsity(adata):
    """
    
    """
    mask = adata.X != 0
    gene_sparsity = np.sum(mask, axis=0) / adata.n_obs
    gene_sparsity = np.asarray(gene_sparsity)
    gene_sparsity = 1 - np.reshape(gene_sparsity, (-1, ))
    adata.var['sparsity'] = gene_sparsity
    
    
def get_matched_genes(prior_genes_names, sn_genes_names, excluded_genes=None):
    """
    Given the list of genes in the spatial data and the list of genes in the single nuclei, identifies the subset of
    genes included in both lists and returns the corresponding matching indices.
    Args:
        prior_genes_names (sequence): List of gene names in the spatial data.
        sn_genes_names (sequence): List of gene names in the single nuclei data.
        excluded_genes (sequence): Optional. List of genes to be excluded. These genes are excluded even if present in both datasets.
            If None, no genes are excluded. Default is None.
    Returns:
        A tuple (mask_prior_indices, mask_sn_indices, selected_genes), with:
            mask_prior_indices (list): List of indices for the selected genes in 'prior_genes_names'.
            mask_sn_indices (list): List of indices for the selected genes in 'sn_genes_names'.
            selected_genes (list): List of names of the selected genes.
        For each i, selected_genes[i] = prior_genes_names[mask_prior_indices[i]] = sn_genes_names[mask_sn_indices[i].
    """
    prior_genes_names = np.array(prior_genes_names)
    sn_genes_names = np.array(sn_genes_names)

    mask_prior_indices = []
    mask_sn_indices = []
    selected_genes = []
    if excluded_genes is None:
        excluded_genes = []
    for index, i in enumerate(sn_genes_names):
        if i in excluded_genes:
            continue
        try:
            mask_prior_indices.append(np.argwhere(prior_genes_names == i)[0][0])
            # if no exceptions above:
            mask_sn_indices.append(index)
            selected_genes.append(i)
        except IndexError:
            pass

    assert len(mask_prior_indices) == len(mask_sn_indices)
    return mask_prior_indices, mask_sn_indices, selected_genes


def one_hot_encoding(l, keep_aggregate=False):
    """
    Given a sequence, returns a DataFrame with a column for each unique value in the sequence and a one-hot-encoding.
    Args:
        l (sequence): List to be transformed.
        keep_aggregate: Optional. If True, the output includes an additional column for the original list. Default is False.
    Returns:
        A DataFrame with a column for each unique value in the sequence and a one-hot-encoding, and an additional
            column with the input list if 'keep_aggregate' is True.
            The number of rows are equal to len(l).
    """
    df_enriched = pd.DataFrame({'cl': l})
    for i in l.unique():
        df_enriched[i] = list(map(int, df_enriched['cl'] == i))
    if not keep_aggregate:
        del df_enriched['cl']
    return df_enriched


def project_cell_annotations(adata_map, annotation='cell_type'):
    """
    Transfer `annotation` from single cell data onto space.
    Args:
        adata_map (AnnData): cell-by-spot AnnData returned by `train` function.
        annotation (str): Cell annotations matrix with shape (number_cells, number_annotations).
    Returns:
        A dataframe with spatial probabilities for each annotation (number_spots, number_annotations)
    """
    df = one_hot_encoding(adata_map.obs[annotation])
    df_ct_prob = adata_map.X.T @ df
    return df_ct_prob


def project_genes(adata_map, adata_sc):
    """
        Transfer gene expression from the single cell onto space.
        Returns a spot-by-gene AnnData containing spatial gene 
        expression from the single cell data.
    """
    if adata_map.obs.index.equals(adata_sc.obs.index) is False:
        raise ValueError('The two AnnDatas need to have same `obs` index.')
    if hasattr(adata_sc.X, 'toarray'):
        adata_sc.X = adata_sc.X.toarray()
    X_space = adata_map.X.T @ adata_sc.X
    adata_ge = sc.AnnData(X=X_space, obs=adata_map.var, var=adata_sc.var)
    training_genes = adata_map.uns['train_genes_df'].index.values
    adata_ge.var['is_training'] = adata_ge.var.index.isin(training_genes)
    return adata_ge


def compare_spatial_geneexp(adata_space_1, adata_space_2):
    """
         Compare gene expression in the two spatial AnnDatas. 
         Used to compared mapped single cell data to original spatial data.
         Returns a DataFrame with similarity scores between genes.
    """
    
    adata_space_1, adata_space_2 = mu.pp_adatas(adata_space_1, adata_space_2)
    annotate_gene_sparsity(adata_space_1)
    annotate_gene_sparsity(adata_space_2)

    # Annotate cosine similarity of each training gene
    cos_sims = []

    if hasattr(adata_space_1.X, 'toarray'):
        X_1 = adata_space_1.X.toarray()
    else:
        X_1 = adata_space_1.X
    if hasattr(adata_space_2.X, 'toarray'):
        X_2 = adata_space_2.X.toarray()
    else:
        X_2 = adata_space_2.X

    for v1, v2 in zip(X_1.T, X_2.T):
        norm_sq = np.linalg.norm(v1) * np.linalg.norm(v2)
        cos_sims.append((v1 @ v2) / norm_sq)
    genes = list(np.reshape(adata_space_1.var.index.values, (-1,)))
    df_g = pd.DataFrame(cos_sims, genes, columns=['score'])
    for adata in [adata_space_1, adata_space_2]:
        if 'is_training' in adata.var.keys():
            df_g['is_training'] = adata.var.is_training
    df_g['sparsity_1'] = adata_space_1.var.sparsity
    df_g['sparsity_2'] = adata_space_2.var.sparsity
    df_g = df_g.sort_values(by='score', ascending=False)
    return df_g


# DEPRECATED
def transfer_annotations_prob(mapping_matrix, to_transfer):
    """
    Transfer cell annotations onto space through a mapping matrix.
    Args:
        mapping_matrix (ndarray): Mapping matrix with shape (number_cells, number_spots).
        to_transfer (ndarray): Cell annotations matrix with shape (number_cells, number_annotations).
    Returns:
        A matrix of annotations onto space, with shape (number_spots, number_annotations)
    """
    return mapping_matrix.transpose() @ to_transfer


def transfer_annotations_prob_filter(mapping_matrix, filter, to_transfer):
    """
    Transfer cell annotations onto space through a mapping matrix and a filter.
    Args:
        mapping_matrix (ndarray): Mapping matrix with shape (number_cells, number_spots).
        filter (ndarray): Filter with shape (number_cells,).
        to_transfer (ndarray): Cell annotations matrix with shape (number_cells, number_annotations).
    Returns:
        A matrix of annotations onto space, with shape (number_spots, number_annotations).
    """
    tt = to_transfer * filter[:, np.newaxis]
    return mapping_matrix.transpose() @ tt


def df_to_cell_types(df, cell_types):
    """
    Utility function that "randomly" assigns cell coordinates in a voxel to known numbers of cell types in that voxel.
    Used for deconvolution.
    Args:
        df (DataFrame): Columns correspond to cell types.  Each row in the DataFrame corresponds to a voxel and
            specifies the known number of cells in that voxel for each cell type (int).
            The additional column 'centroids' specifies the coordinates of the cells in the voxel (sequence of (x,y) pairs).
        cell_types (sequence): Sequence of cell type names to be considered for deconvolution.
            Columns in 'df' not included in 'cell_types' are ignored for assignment.
    Returns:
        A dictionary <cell type name> -> <list of (x,y) coordinates for the cell type>
    """
    df_cum_sums = df[cell_types].cumsum(axis=1)

    df_c = df.copy()

    for i in df_cum_sums.columns:
        df_c[i] = df_cum_sums[i]

    cell_types_mapped = defaultdict(list)
    for i_index, i in enumerate(cell_types):
        for j_index, j in df_c.iterrows():
            start_ind = 0 if i_index == 0 else j[cell_types[i_index - 1]]
            end_ind = j[i]
            cell_types_mapped[i].extend(j['centroids'][start_ind:end_ind].tolist())
    return cell_types_mapped



"""
Library for instantiating and running the optimizer for Tangram. The optimizer comes in two flavors,
which correspond to two different classes:
- Mapper: optimizer without filtering (i.e., all single cells are mapped onto space). At the end, the learned mapping
matrix M is returned.
- MapperConstrained: optimizer with filtering (i.e., only a subset of single cells are mapped onto space).
At the end, the learned mapping matrix M and the learned filter F are returned.
"""
import numpy as np
import logging
import torch
from torch.nn.functional import softmax, cosine_similarity


class Mapper:
    """
    Allows instantiating and running the optimizer for Tangram, without filtering.
    Once instantiated, the optimizer is run with the 'train' method, which also returns the mapping result.
    """
    def __init__(self, S, G, d, 
                 lambda_g1=1., lambda_d=0, lambda_g2=0, lambda_r=0, device='cpu', adata_map=None):
        """
        Instantiate the Tangram optimizer (without filtering).
        Args:
            S (ndarray): Single nuclei matrix, shape = (number_cell, number_genes).
            G (ndarray): Spatial transcriptomics matrix, shape = (number_spots, number_genes).
                Spots can be single cells or they can contain multiple cells.
            d (ndarray): Spatial density of cells, shape = (number_spots,).
                This array should satisfy the constraints d.sum() == 1.
            lambda_g1 (float): Optional. Strength of Tangram loss function. Default is 1.
            lambda_d (float): Optional. Strength of density regularizer. Default is 0.
            lambda_g2 (float): Optional. Strength of voxel-gene regularizer. Default is 0.
            lambda_r (float): Optional. Strength of entropy regularizer. An higher entropy promotes
                probabilities of each cell peaked over a narrow portion of space.
                lambda_r = 0 corresponds to no entropy regularizer. Default is 0.
            device (str or torch.device): Optional. Device is 'cpu'.
            adata_map (scanpy.AnnData): Optional. Mapping initial condition (for resuming previous mappings)
        """
        self.S = torch.tensor(S, device=device, dtype=torch.float32)
        self.G = torch.tensor(G, device=device, dtype=torch.float32)
        self.d = torch.tensor(d, device=device, dtype=torch.float32)

        self.lambda_d = lambda_d
        self.lambda_g1 = lambda_g1
        self.lambda_g2 = lambda_g2
        self.lambda_r = lambda_r
        self._density_criterion = torch.nn.KLDivLoss(reduction='sum')
        
        if adata_map is None:
            self.M = np.random.normal(0, 1, (S.shape[0], G.shape[0]))
        else:
            raise NotImplemented
            self.M = adata_map.X  # doesn't work. maybe apply inverse softmax
            
        self.M = torch.tensor(self.M, device=device, requires_grad=True, dtype=torch.float32)

    def _loss_fn(self, verbose=True):
        """
        Evaluates the loss function.
        Args:
            verbose (bool): Optional. Whether to print the loss results. If True, the loss for each individual term is printed as:
                density_term, gene-voxel similarity term, voxel-gene similarity term. Default is True.
        Returns:
            Total loss (float).
        """
        M_probs = softmax(self.M, dim=1)

        d_pred = torch.log(M_probs.sum(axis=0) / self.M.shape[0])  # KL wants the log in first argument
        density_term = self.lambda_d * self._density_criterion(d_pred, self.d)

        G_pred = torch.matmul(M_probs.t(), self.S)
        gv_term = self.lambda_g1 * cosine_similarity(G_pred, self.G, dim=0).mean()
        vg_term = self.lambda_g2 * cosine_similarity(G_pred, self.G, dim=1).mean()
        expression_term = gv_term + vg_term

        regularizer_term = self.lambda_r * (torch.log(M_probs) * M_probs).sum()

        if verbose:
            main_loss = (gv_term / self.lambda_g1).tolist()
            kl_reg = (density_term / self.lambda_d).tolist()
            vg_reg = (vg_term / self.lambda_g2).tolist()
            msg = 'Score: {:.3f}, KL reg: {:.3f}, VG reg: {:.3f}'.format(
                main_loss, kl_reg, vg_reg
            )
            print(msg)
            
        return density_term - expression_term - regularizer_term

    def train(self, num_epochs, learning_rate=0.1, print_each=100):
        """
        Run the optimizer and returns the mapping outcome.
        Args:
            num_epochs (int): Number of epochs.
            learning_rate (float): Optional. Learning rate for the optimizer. Default is 0.1.
            print_each (int): Optional. Prints the loss each print_each epochs. If None, the loss is never printed. Default is 100.
        Returns:
            The optimized mapping matrix M (ndarray), with shape (number_cells, number_spots).
        """
        optimizer = torch.optim.Adam([self.M], lr=learning_rate)
        logging.warning(f'Printing scores every {print_each} epochs.')
        for t in range(num_epochs):
            if print_each is None or t % print_each != 0:
                loss = self._loss_fn(verbose=False)
            else:
                loss = self._loss_fn(verbose=True)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        # take final softmax w/o computing gradients
        with torch.no_grad():
            output = softmax(self.M, dim=1).cpu().numpy()
            return output


class MapperConstrained:
    """
    Allows instantiating and running the optimizer for Tangram, with filtering.
    Once instantiated, the optimizer is run with the 'train' method, which also returns the mapping and filter results.
    """
    def __init__(self, S, G, d, lambda_d=1, lambda_g1=1, lambda_g2=1, lambda_r=0, lambda_count=1, lambda_f_reg=1,
                 device='cpu', target_count=None):
        """
        Instantiate the Tangram optimizer (with filtering).
        Args:
            S (ndarray): Single nuclei matrix, shape = (number_cell, number_genes).
            G (ndarray): Spatial transcriptomics matrix, shape = (number_spots, number_genes).
                Spots can be single cells or they can contain multiple cells.
            d (ndarray): Spatial density of cells, shape = (number_spots,).
                This array should satisfy the constraints d.sum() == 1.
            lambda_d (float): Optional. Hiperparameter for the density term of the optimizer. Default is 1.
            lambda_g1 (float): Optional. Hyperparameter for the gene-voxel similarity term of the optimizer. Default is 1.
            lambda_g2 (float): Optional. Hyperparameter for the voxel-gene similarity term of the optimizer. Default is 1.
            lambda_r (float): Optional. Entropy regularizer for the learned mapping matrix. An higher entropy promotes
                probabilities of each cell peaked over a narrow portion of space.
                lambda_r = 0 corresponds to no entropy regularizer. Default is 0.
            lambda_count (float): Optional. Regularizer for the count term. Default is 1.
            lambda_f_reg (float): Optional. Regularizer for the filter, which promotes Boolean values (0s and 1s) in the filter. Default is 1.
            target_count (int): Optional. The number of cells to be filtered. If None, this number defaults to the number of
                voxels inferred by the matrix 'G'. Default is None.
        """
        self.S = torch.tensor(S, device=device, dtype=torch.float32)
        self.G = torch.tensor(G, device=device, dtype=torch.float32)
        self.d = torch.tensor(d, device=device, dtype=torch.float32)

        self.lambda_d = lambda_d
        self.lambda_g1 = lambda_g1
        self.lambda_g2 = lambda_g2
        self.lambda_r = lambda_r
        self.lambda_count = lambda_count
        self.lambda_f_reg = lambda_f_reg
        self._density_criterion = torch.nn.KLDivLoss(reduction='sum')

        if target_count is None:
            self.target_count = self.G.shape[0]
        else:
            self.target_count = target_count

        self.M = np.random.normal(0, 1, (S.shape[0], G.shape[0]))
        self.M = torch.tensor(self.M, device=device, requires_grad=True, dtype=torch.float32)

        self.F = np.random.normal(0, 1, S.shape[0])
        self.F = torch.tensor(self.F, device=device, requires_grad=True, dtype=torch.float32)

    def _loss_fn(self, verbose=True):
        """
        Evaluates the loss function.
        Args:
            verbose (bool): Optional. Whether to print the loss results. If True, the loss for each individual term is printed as:
                density_term, gene-voxel similarity term, voxel-gene similarity term. Default is True.
        Returns:
            Total loss (float).
        """
        M_probs = softmax(self.M, dim=1)
        F_probs = torch.sigmoid(self.F)

        M_probs_filtered = M_probs * F_probs[:, np.newaxis]
        d_pred = torch.log(M_probs_filtered.sum(axis=0) / (F_probs.sum()))  # KL wants the log in first argument
        density_term = self.lambda_d * self._density_criterion(d_pred, self.d)

        S_filtered = self.S * F_probs[:, np.newaxis]

        G_pred = torch.matmul(M_probs.t(), S_filtered)
        gv_term = self.lambda_g1 * cosine_similarity(G_pred, self.G, dim=0).mean()
        vg_term = self.lambda_g2 * cosine_similarity(G_pred, self.G, dim=1).mean()
        expression_term = gv_term + vg_term

        regularizer_term = self.lambda_r * (torch.log(M_probs) * M_probs).sum()

        _count_term = F_probs.sum() - self.target_count
        count_term = self.lambda_count * torch.abs(_count_term)

        f_reg_t = F_probs - F_probs * F_probs
        f_reg = self.lambda_f_reg * f_reg_t.sum()

        if verbose:
            print((density_term / self.lambda_d).tolist(),
                  (gv_term / self.lambda_g1).tolist(),
                  (vg_term / self.lambda_g2).tolist(),
                  (count_term / self.lambda_count).tolist(),
                  (f_reg / self.lambda_f_reg).tolist()
                  )

        return density_term - expression_term - regularizer_term + count_term + f_reg

    def train(self, num_epochs, learning_rate=0.1, print_each=100):
        """
        Run the optimizer and returns the mapping outcome.
        Args:
            num_epochs (int): Number of epochs.
            learning_rate (float): Optional. Learning rate for the optimizer. Default is 0.1.
            print_each (int): Optional. Prints the loss each print_each epochs. If None, the loss is never printed. Default is 100.
        Returns:
            A tuple (M, f), with:
                M (ndarray) is the optimized mapping matrix, shape = (number_cells, number_spots).
                f (ndarray) is the optimized filter, shape = (number_cells,).
        """
        optimizer = torch.optim.Adam([self.M, self.F], lr=learning_rate)

        for t in range(num_epochs):
            if print_each is None or t % print_each != 0:
                loss = self._loss_fn(verbose=False)
            else:
                loss = self._loss_fn(verbose=True)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
        # take final softmax w/o computing gradients
        with torch.no_grad():
            output = softmax(self.M, dim=1).cpu().numpy()
            F_out = torch.sigmoid(self.F).cpu().numpy()
            return output, F_out



"""
This module includes plotting utility functions.
"""
import numpy as np
import matplotlib.pyplot as plt
import logging
import seaborn as sns
from scipy.stats import entropy
import scanpy as sc
from scipy.sparse.csc import csc_matrix
from scipy.sparse.csr import csr_matrix

from . import utils as ut
from . import mapping_utils as mu


def plot_training_scores(adata_map, bins='auto', alpha=.7):
    """
    
    """
    fig, axs = plt.subplots(1, 4, figsize=(12, 3), sharey=True)
    df = adata_map.uns['train_genes_df']
    axs_f = axs.flatten()
    
#     axs_f[0].set_title('Training scores for single genes')
    sns.histplot(data=df, y='train_score', bins=10, ax=axs_f[0]);

    axs_f[1].set_title('score vs sparsity (single cells)')
    sns.scatterplot(data=df, y='train_score', x='sparsity_sc', ax=axs_f[1], alpha=alpha)
#     sns.distplot(data=df, x='train_score', bins=bins, ax=axs_f[0]);
    
    axs_f[2].set_title('score vs sparsity (spatial)')
    sns.scatterplot(data=df, y='train_score', x='sparsity_sp', ax=axs_f[2], alpha=alpha)
    
    axs_f[3].set_title('score vs sparsity (sp - sc)')
    sns.scatterplot(data=df, y='train_score', x='sparsity_diff', ax=axs_f[3], alpha=alpha)
    
    plt.tight_layout()

    
def plot_gene_sparsity(adata_1, adata_2, xlabel='adata_1', ylabel='adata_2', s=1, genes=None):
    """
        Compare sparsity of all genes between `adata_1` and `adata_2`.
    """
    logging.info('Pre-processing AnnDatas...')
    adata_1, adata_2 = mu.pp_adatas(adata_1, adata_2, genes=genes)
    logging.info('Annotating sparsity...')
    ut.annotate_gene_sparsity(adata_1)
    ut.annotate_gene_sparsity(adata_2)
    xs = adata_1.var['sparsity'].values
    ys = adata_2.var['sparsity'].values
    fig, ax = plt.subplots(1, 1)
    ax.set_aspect(1)
    ax.set_xlabel('sparsity (' + xlabel + ')')
    ax.set_ylabel('sparsity (' + ylabel + ')')
    ax.set_title(f'Gene sparsity ({len(xs)} genes)')
    ax.scatter(xs, ys, s=s, marker='x')

    
def ordered_predictions(xs, ys, preds, reverse=False):
    """
    Utility function that orders 2d points based on values associated to each point.
    Args:
        xs: Sequence of x coordinates (floats).
        ys: Sequence of y coordinates (floats).
        ys: Sequence of y coordinates (floats).
        reverse: Optional. A Boolean. False will sort ascending, True will sort descending. Default is False.
    Returns:
        Returns the ordered xs, ys, preds.
    """
    assert len(xs) == len(ys) == len(preds)
    return list(zip(*[(x, y, z) for x, y, z in sorted(zip(xs, ys, preds), key=lambda pair: pair[2], reverse=reverse)]))


def plot_cell_annotation(adata_map, annotation='cell_type', 
                         x='x', y='y', nrows=None, ncols=None,
                         marker_size=5, cmap='viridis', suptitle_add=False):
    """
        Transfer an annotation for a single cell dataset onto space, and visualize
        corresponding spatial probability maps.
        `adata_map`: cell-by-spot-AnnData containing mapping result
        `annotation`: Must be a column in `adata_map.obs`.
        `x`: column name for spots x-coordinates (must be in `adata_map.var`)
        `y`: column name for spots y-coordinates (must be in `adata_map.var`)
    """

    # TODO ADD CHECKS for x and y
    
    df_annotation = ut.project_cell_annotations(adata_map, annotation=annotation)

    if nrows is None or ncols is None:
        ncols = 1
        nrows = len(df_annotation.columns)

    fig, axs = plt.subplots(nrows, ncols, 
                              figsize=(ncols*3, nrows*3), 
                              sharex=True, sharey=True) 

    axs_f = axs.flatten()
    [ax.axis('off') for ax in axs_f]

    if len(df_annotation.columns) > nrows*ncols:
        logging.warning('Number of panels smaller than annotations. Increase `nrows`/`ncols`.')
    
    iterator = zip(df_annotation.columns, range(nrows*ncols))
    for ann, index in iterator:
        xs, ys, preds = ordered_predictions(adata_map.var[x], 
                                            adata_map.var[y], 
                                            df_annotation[ann])
        axs_f[index].scatter(x=xs, y=ys, c=preds, s=marker_size, cmap=cmap)
#         axs_f[index].axis('off')
        axs_f[index].set_aspect(1)
        axs_f[index].set_title(ann)
        
    if suptitle_add is True:
        fig.suptitle(annotation)


def plot_genes(genes, adata_measured, adata_predicted, x='x', y='y', s=5, log=False):
    """

    """
    # TODO: not very elegant and slow as hell
    if isinstance(adata_measured.X, csc_matrix) or isinstance(adata_measured.X, csr_matrix):
        adata_measured.X = adata_measured.X.toarray()
    
    fig, axs = plt.subplots(nrows=len(genes), ncols=2, figsize=(6, len(genes)*3))
    for ix, gene in enumerate(genes):
        xs, ys, vs = ordered_predictions(adata_measured.obs[x], 
                                         adata_measured.obs[y], 
                                         np.array(adata_measured[:, gene].X).flatten())
        if log:
            vs = np.log(1+np.asarray(vs))
        axs[ix, 0].scatter(xs, ys, c=vs, cmap='inferno', s=s)
        axs[ix, 0].set_title(gene + ' (measured)')
        axs[ix, 0].axis('off')
        
        xs, ys, vs = ordered_predictions(adata_predicted.obs[x], 
                                         adata_predicted.obs[y], 
                                         np.array(adata_predicted[:, gene].X).flatten())
        if log:
            vs = np.log(1+np.asarray(vs))
        axs[ix, 1].scatter(xs, ys, c=vs, cmap='inferno', s=s)
        axs[ix, 1].set_title(gene + ' (predicted)')
        axs[ix, 1].axis('off')
    
    
def quick_plot_gene(gene, adata, x='x', y='y', s=50, log=False):
    """
    Utility function to quickly plot a gene in a AnnData structure ordered by intensity of the gene signal.
    Args:
        gene (str): Gene name.
        adata: AnnData structure.
        x: Optional. Name for the first coordinate in AnnData.obs. Default is 'x'.
        y: Optional. Name for the second coordinate in AnnData.obs. Default is 'y'.
        log: Optional. Whether to apply the log before plotting. Default is False.
    """
    xs, ys, vs = ordered_predictions(adata.obs[x], adata.obs[y], np.array(adata[:, gene].X).flatten())
    if log:
        vs = np.log(1+np.asarray(vs))
    plt.scatter(xs, ys, c=vs, cmap='viridis', s=s)


def plot_annotation_entropy(adata_map, annotation='cell_type'):
    """
        
    """
    qk = np.ones(shape=(adata_map.n_obs, adata_map.n_vars))
    adata_map.obs['entropy'] = entropy(adata_map.X, base=adata_map.X.shape[1], axis=1)
    fig, ax = plt.subplots(1, 1, figsize=(10, 3))
    ax.set_ylim(0, 1)
    sns.boxenplot(x=annotation, y="entropy", data=adata_map.obs, ax=ax);
    plt.xticks(rotation=30);


# Colors used in the manuscript for deterministic assignment.
mapping_colors = {'L6 CT': (0.19215686274509805, 0.5098039215686274, 0.7411764705882353),
                  'L6 IT': (0.4196078431372549, 0.6823529411764706, 0.8392156862745098),
                  'L5/6 NP': (0.6196078431372549, 0.792156862745098, 0.8823529411764706),
                  'L6b': '#0000c2ff',
                  'L2/3 IT': (0.9019607843137255, 0.3333333333333333, 0.050980392156862744),
                  'L5 IT': (0.19215686274509805, 0.6392156862745098, 0.32941176470588235),
                  'L5 ET': (0.4549019607843137, 0.7686274509803922, 0.4627450980392157),
                  'Oligo': (0.4588235294117647, 0.4196078431372549, 0.6941176470588235),
                  'Vip': (0.6196078431372549, 0.6039215686274509, 0.7843137254901961),
                  'Astro': '#ffdd55ff',
                  'Micro-PVM': '#000000ff',
                  'Pvalb': (0.38823529411764707, 0.38823529411764707, 0.38823529411764707),
                  'Lamp5': (0.5882352941176471, 0.5882352941176471, 0.5882352941176471),
                  'Sst': (0.7411764705882353, 0.7411764705882353, 0.7411764705882353),
                  'Sst Chodl': (0.8509803921568627, 0.8509803921568627, 0.8509803921568627),
                  'Sncg': (0.5176470588235295, 0.23529411764705882, 0.2235294117647059),
                  'Peri': (0.6784313725490196, 0.28627450980392155, 0.2901960784313726),
                  'VLMC': (0.8392156862745098, 0.3803921568627451, 0.4196078431372549),
                  'Endo': (0.9058823529411765, 0.5882352941176471, 0.611764705882353),
                  'Meis2': '#FFA500ff',
                  'SMC': '#000000ff',
                  'L6 PT': '#4682B4ff',
                  'L5 PT': '#a1ed7bff',
                  'L5 NP': '#6B8E23ff',
                  'L4': '#d61f1dff',
                  'Macrophage': '#2b2d2fff',
                  'CR': '#000000ff'}


#!/usr/bin/env python3

import setuptools

with open("README.md","r", encoding="utf-8") as fh:
    long_description = fh.read()

d = {}
with open('tangram/_version.py') as f:  exec(f.read(), d)

setuptools.setup(
    name="tangram-sc",
    version=d['__version__'],
    author="Tommaso Biancalani, Gabriele Scalia",
    author_email="tommaso.biancalani@gmail.com",
    description="Spatial alignment of single cell transcriptomic data.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/broadinstitute/Tangram",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Operating System :: MacOS",
    ],
    python_requires='>=3.6',
    install_requires=[
        "pip>=19.0.0",
        "torch>=1.4.0",
        "pandas>=1.1.0",
        "numpy>=1.19.1",
        "scipy>=1.5.2",
        "matplotlib>=3.0.0",
        "seaborn>=0.10.1",
        "scanpy==1.6.0",
        # "jupyterlab>=2.2.6",
    ]
)



